package org.basex.modules.ftp;

import java.io.*;

import org.apache.commons.net.ftp.*;
import org.basex.io.in.*;
import org.basex.io.out.*;
import org.basex.query.*;
import org.basex.query.value.*;
import org.basex.query.value.item.*;
import org.basex.util.*;

/**
 * FTP Connections.
 *
 * @author BaseX Team 2005-12, BSD License
 * @author Christian Gruen
 */
public final class FTP extends QueryModule {
  /** FTP client. */
  final FTPClient ftp = new FTPClient();

  /**
   * Connects to an FTP server, using the standard port 21.
   * @param server server path
   * @param user user name
   * @param pass password
   * @throws QueryException query exception
   */
  public void connect(final Str server, final Str user, final Str pass)
      throws QueryException {
    connect(server, Int.get(21), user, pass);
  }

  /**
   * Connects to an FTP server.
   * @param server server path
   * @param port port
   * @param user user name
   * @param pass password
   * @throws QueryException query exception
   */
  public void connect(final Str server, final Int port, final Str user, final Str pass)
      throws QueryException {

    try {
      ftp.connect(server.toJava(), (int) port.itr());
      if(!FTPReply.isPositiveCompletion(ftp.getReplyCode()))
        throw new Exception("Could not connect.");

      if(!ftp.login(user.toJava(), pass.toJava())) {
        ftp.logout();
        throw new Exception("Access denied.");
      }
    } catch(final Exception ex) {
      throw error(ex);
    }
  }

  /**
   * Disconnects from an FTP server.
   * @throws QueryException query exception
   */
  public void disconnect() throws QueryException {
    try {
      ftp.disconnect();
    } catch(final Exception ex) {
      throw error(ex);
    }
  }

  /**
   * Lists the file names of a remote FTP directory.
   * @param path remote path
   * @return files
   * @throws QueryException query exception
   */
  public Value list(final Str path) throws QueryException {
    final ValueBuilder vb = new ValueBuilder();
    try {
      for(final FTPFile f : ftp.listFiles(path.toJava())) {
        vb.add(Str.get(f.getName()));
      }
      return vb.value();
    } catch(final Exception ex) {
      throw error(ex);
    }
  }

  /**
   * Retrieves the specified file as text.
   * @param path remote path
   * @return text
   * @throws QueryException query exception
   */
  public Str getText(final Str path) throws QueryException {
    return Str.get(get(path));
  }

  /**
   * Retrieves the specified file as Base64 binary.
   * @param path remote path
   * @return text
   * @throws QueryException query exception
   */
  public B64 getBinary(final Str path) throws QueryException {
    return B64.get(get(path));
  }

  /**
   * Uploads a string to the FTP server.
   * @param path remote path
   * @param string textual string to be uploaded
   * @throws QueryException query exception
   */
  public void putText(final Str path, final Str string) throws QueryException {
    put(path, new ArrayInput(string.string()));
  }

  /**
   * Uploads a binary item to the FTP server.
   * @param path remote path
   * @param it binary item to be uploaded
   * @throws QueryException query exception
   */
  public void putBinary(final Str path, final Bin it) throws QueryException {
    put(path, it.input(null));
  }

  /**
   * Deletes a file from the FTP server.
   * @param path remote file name
   * @return true if successfully completed, false if not.
   * @throws QueryException query exception
   */
  public Bln delete(final Str path) throws QueryException {
    try {
      return Bln.get(ftp.deleteFile(path.toJava()));
    } catch(final Exception ex) {
      throw error(ex);
    }
  }

  /**
   * Retrieves the specified file.
   * @param path remote path
   * @return file contents
   * @throws QueryException query exception
   */
  private byte[] get(final Str path) throws QueryException {
    try {
      final ArrayOutput ao = new ArrayOutput();
      ftp.retrieveFile(path.toJava(), ao);
      return ao.toArray();
    } catch(final Exception ex) {
      throw error(ex);
    }
  }

  /**
   * Stores the specified file.
   * @param path remote path
   * @param input data to be stored
   * @throws QueryException query exception
   */
  private void put(final Str path, final InputStream input) throws QueryException {
    try {
      try {
        ftp.storeFile(path.toJava(), input);
      } finally {
        input.close();
      }
    } catch(final Exception ex) {
      throw error(ex);
    }
  }

  /**
   * Throws a query exception for the specified exception.
   * @param ex exception
   * @return query exception
   * @throws QueryException query exception
   */
  private static QueryException error(final Exception ex) throws QueryException {
    throw new QueryException("FTP: " + Util.message(ex));
  }
}
